%------------------------------------------------------------%
%--- License terms ------------------------------------------%
%------------------------------------------------------------%

%--- This program is free software: you can redistribute it 
%--- and/or modify it under the terms of the GNU General Public 
%--- License as published by the Free Software Foundation, either 
%--- version 3 of the License, or (at your option) any later 
%--- version. This program is distributed in the hope that it 
%--- will be useful, but without any warranty; without even the 
%--- implied warranty of merchantability or fitness for a
%--- particular purpose. See the GNU General Public License for 
%--- more details. You should have received a copy of the
%--- GNU General Public License along with this program; 
%--- if not, write to the Free Software Foundation, Inc., 
%--- 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335, 
%--- USA. 

%------------------------------------------------------------%
%--- TV regularization after Chambolle/Pock (2011) ----------%
%------------------------------------------------------------%

function [u e] = tgv1_l2_2D_pd(g, f, alpha0, maxits)

check_it = 10;
[M N] = size(f);
xi = zeros(M, N, 2);
u = f;
u_ = zeros(M, N);
e = [];

%------------------------------------------------------------%
%--- Note: multiindices of the spatial derivatives are ------%
%------------------------------------------------------------%

% derivatives 
% | uxx uxy |
% | uyx uyy |

% multiindices
% | 1 3 |
% | 3 2 |

%------------------------------------------------------------%
%------------------------------------------------------------%

L = sqrt(8);
tau_p = 1/L;
tau_d = 1/L;

%=== Main loop ==============================================%

for k = 0 : maxits,

%------------------------------------------------------------%  
%--- Primal update ------------------------------------------%
%------------------------------------------------------------%
  
div = dxm(xi(:,:,1)) + dym(xi(:,:,2));
  
%--- remember previous value of u

u_ = -u;
  
%--- primal update

u = u + tau_p * div;
  
%--- proximal operator

u = (u + tau_p * f) ./ (1.0 + tau_p);
  
%--- overrelaxation

u_ = u_ + 2.0 * u;

%------------------------------------------------------------%  
%--- Dual update --------------------------------------------%
%------------------------------------------------------------%
  
u_x = dxp(u_); 
u_y = dyp(u_);
xi(:,:,1) = xi(:,:,1) + tau_d * u_x;
xi(:,:,2) = xi(:,:,2) + tau_d * u_y;

%------------------------------------------------------------%  
%--- Reprojection--------------------------------------------%
%------------------------------------------------------------%
 
%--- compute the minimizer of
%--- min_{xi} || xi - xi_old ||^2  s.t. || xi || <= alpha0

%--- initialization
  
denom = max(1.0, sqrt(xi(:,:,1).^2 + xi(:,:,2).^2) / alpha0);
xi(:,:,1) = xi(:,:,1) ./ denom;
xi(:,:,2) = xi(:,:,2) ./ denom;
 
if mod(k, check_it) == 0,
div = dxm(xi(:,:,1)) + dym(xi(:,:,2));
primal_dual_energy = -sum(sum(u .* div)) + 0.5 * sum(sum((u-f).^2));
u_ = f + div;
dual_energy = sum(sum(u.^2));
e = [e dual_energy];
rmse = sqrt(mean(mean((u-g).^2)));
fprintf('TGV1-L2-2D-PD: it = %4d,  e = %f, rmse = %f\n', k, e(end), rmse);
end

%=== End of main loop =======================================%  
end

end

function [dx] = dxm(u)
[M N] = size(u);
dx = [u(:, 1:(end-1)) zeros(M,1)] - [zeros(M,1) u(:, 1:(end-1))];
end

function [dx] = dxp(u)
[M N] = size(u);
dx = [u(:, 2:end) u(:, end)] - u;
end

function [dy] = dym(u)
[M N] = size(u);
dy = [u(1:(end-1),:); zeros(1,N)] - [zeros(1,N); u(1:(end-1),:)];
end

function [dy] = dyp(u)
[M N] = size(u);
dy = [u(2:end, :); u(end, :)] - u;
end